#!/usr/bin/perl -wl

$version = '0.1.1';

# Include-file resides in same directory as program itself
push(@INC,$0 =~ /(.*\/)/);

require 'qfits.pl';

defined($qfits::version) || die;
defined($qfits::support) || die;

$help = "
  $qfits::prog version $version (using qfits.pl version $qfits::version).
  Print FITS-header cards to stdout.

  Proto-type: Options may change in future versions.

Usage:

  $qfits::prog [options] 1.fits [2.fits [3.fits []]]

The file '-' is allowed once and is interpreted as standard input.

Options:
  -x  n                     : Print FITS-header extension n. n = 0 means all
                              headers.
  -d [start[[:step]:stop]]  : Do not print any header-cards, instead print
                              the numbers in the Data Unit. Numbering starts
                              with 1, no numbering prints all numbers. With -d
                              a number is defined by the headers BITPIX value.
                              Table data is therefore always printed as
                              unsigned 8-bit integers. See also dtfits.pl.
  -l n                      : Print the file position of the FITS-headers in
                              extensions 1 through n. n = 0 means all headers.
                              Print also the size of the (zero-padded) headers
                              and data units.
  -b                        : Ignore blank cards. Implies absence of -d.
  -verbose                  : Print some extra information to stderr.
  -help                     : Print this message to stderr and exit.

Example:

  $qfits::prog -x 1 my.tfits

Bugs: With '-' the response time may be longer and detection of truncated input
may fail.

For further support contact $qfits::support

";


@ARGV || die $help;

$dolist = -1;
$doext = -1;
$verbose = 0;
$dodata = 0;
$noblank = 0;
for (;;) {

  $ARGV[0] eq '-l' && $ARGV[1] =~ /^\d+$/ and shift(), $dolist = 0+shift(), next;
  $ARGV[0] eq '-x' && $ARGV[1] =~ /^\d+$/ and shift(), $doext = 0+shift(), next;
  $ARGV[0] eq '-b' and shift(), $noblank = 1, next;
  if ($ARGV[0] eq '-d') {
    $dodata = 1;
    shift();
    $stop = -1;
    $start = $step = 1;
    $ARGV[0] =~ /^(\d+)(?:(?:\:(\d+))?\:(\d+))?$/ && (!$3 || $3 >= $1) and shift(),
      ($start,$step,$stop) = ($1, $2 || 1,$3 || -1);
    next;
  }
  $ARGV[0] eq '-verbose' and shift(), $verbose = 1, next;
  $ARGV[0] eq '-help' and warn($help), exit;

  last;

}

$noblank && $dodata and die $help;
$dolist >=0 and $dodata || $doext >= 0 and die $help;
@ARGV || die $help;

$dolist >=0 and $doext = $dolist, $filepos = 0;

for $file (@ARGV) {

  print "====> file $file (main) <====";

  local(*IN);
  if ($file eq '-') {
    *IN = *STDIN;
  } else {
    open(IN,$file) or warn("error: cannot open file [$file]\n"), next;
  }

  # Count number of extensions
  for ($mext = 0; ; $mext++) {

    $doext > 0 && $doext == $mext || $doext == 0 && $mext > 0  ||
      $dolist >= 0 && $mext > 0 and
      print "====> xtension $mext";

    if ($doext == 0 || $doext == $mext || $mext == 0 || $dolist >=0) {

      $dolist >=0 and $filepos = tell(IN);
      local($header_size) = &header(*IN,\@cards) or last;

      if ($mext == 0) {
        $cards[0] =~ /^SIMPLE  =/ or warn("Not a FITS file: Filepos=$filepos ",
        "has data that is not the beginning of a FITS extension header\n"), last;
      } else {
        $cards[0] =~ /^XTENSION=/ or warn("Not a FITS file: Filepos=$filepos ",
        "has data that is not the beginning of a FITS extension header\n"), last;
      }

      &split_cards(\@cards,\%vals,\%comments);
      %comments = ();
      local($data_size) = &data_size(\%vals);

      if ($doext == 0 || $doext == $mext || $mext == 0 && $doext < 0 ||
          $dolist >= 0) {
        # Print only if requested

        if ($dodata) {
          &dodata(*IN,$start,$step,$stop) ? next : last;
	} elsif ($dolist >=0) {
          $mext and print("Filepos=$filepos");
          print("Headersize=$header_size");
          print("Datasize=$data_size");
	} else {
          grep($noblank && /^ +$/ || print(),@cards);
        }

        $dolist < 0 && $doext || $dolist > 0 && $mext == $dolist and last;
      }

      &skip_data(*IN, $data_size) || last;

      $mext==0 and $doext < 0 || !defined($vals{EXTEND}) || $vals{EXTEND} ne 'T'
        and last;

    } else {
      &skip_xt_hdu(*IN) || last;
    }

    eof(IN) && last;

  }
}

$verbose && warn(&run_stats);


sub dodata {
  local(*IN,$start,$step,$stop) = @_;

  local($octets) = abs($vals{BITPIX})>>3;
  local($nd) = 2880 / $octets;
  $nd == 360 || $nd == 720 || $nd == 1440 || $nd == 2880 or
    warn("Illegal octet-length: $nd\n"), return;

  local($nums) = data_entries(\%vals);
  local($size) = &data_size(\%vals);
  local($rpos) = 0;

  $start > $nums and warn("Data Unit contains only $nums entries\n"), return;

  local($i) = $start-1;

  $i -= $i % $nd;
  if ($i) {
    local($seek) = $octets * $i;
    $seek > $size and warn("Seek beyond Data Unit\n"), return;
    &skip_data(*IN,$seek) or return;
    $rpos = $seek;
  }

  local(@nums)=();

  for (;$stop < 0 || $i < $stop and $i < $nums; $i++) {
    @nums or $rpos >= $size && last, $rpos+=2880,
      &read_data_record(*IN, \@nums, $vals{BITPIX}) || last;
    $_ = shift(@nums);
    $i >= $start -1 && $i % $step == 0 || next;
    defined($vals{BSCALE}) and $_ *= $vals{BSCALE};
    defined($vals{BZERO})  and $_ += $vals{BZERO};
    print;
  }

}
