#!/usr/bin/perl -w

$version = '0.1.1';

# Include-file resides in same directory as program itself
push(@INC,$0 =~ /(.*\/)/);

require 'qfits.pl';

defined($qfits::version) || die;
defined($qfits::support) || die;

defined(%qfits::tformbitpix) || die;
defined(%qfits::tform) || die;
defined($qfits::mtform) || die;
defined($qfits::doswap) || die;
defined(%qfits::tformunpack) || die;
defined(%qfits::tformpwidth) || die;
defined(%qfits::tformpwidth0) || die;

$help = "
  $qfits::prog version $version (using qfits.pl version $qfits::version).
  Print the content of a binary FITS-table to stdout as ASCII.

  Proto-type: Options may change in future versions.

Usage:

  $qfits::prog [options] 1.tfits [2.tfits [3.tfits []]]

The file '-' is allowed once and is interpreted as standard input.

Options:
  -d         : Print only table contents (no header lines)
  -s <char>  : Separate the columns with <char>
  -verbose   : Print some extra information to stderr.
  -help      : Print this message to stderr and exit.

Example:

  $qfits::prog all_my*.{,t}fits

Bugs: With '-' the output will display a '?' for the number of
extensions and the response time may be longer.

For further support contact $qfits::support

";

$verbose = 0;

@ARGV || die $help;

$sep = '|';
$head = 1;
for (;;) {

  $ARGV[0] eq '-d' and shift(), $head = 0, next;
  $ARGV[0] eq '-s' and shift(), $sep = shift() || die($help), next;
  $ARGV[0] eq '-verbose' and shift(), $verbose = 1, next;
  $ARGV[0] eq '-help' and warn($help), exit;

  last;

}


for $file (@ARGV) {

  local(*IN);
  if ($file eq '-') {
    *IN = *STDIN;
    $mext = '?';
  } else {

    open(IN,$file) or warn("error: cannot open file [$file]\n"), next;

    # Count number of extensions
    for ($mext = -1; !eof(IN); $mext++) {

      $mext >= 0 and &skip_xt_hdu(*IN), next;

      # The primary header must be checked for EXTEND = T

      &header(*IN,\@cards);

      &split_cards(\@cards,\%vals,\%comments);

      &skip_data(*IN, &data_size(\%vals));

      $vals{EXTEND} ne 'T' and eof(IN) ||
        warn("$file has ".((-s $file) - tell(IN))." trailing bytes\n"),
        last;

    }
    # There are $mext extensions

    seek(IN,0,0);

  }

  for ($ext = 0; $file eq '-' ? !eof(IN) : $ext <= $mext; $ext++) {

    &header(*IN,\@cards);

    &split_cards(\@cards,\%vals,\%comments);

    local($size) = &data_size(\%vals);

    if (!defined($vals{XTENSION})) {
        $size and warn('Skipping '.($ext ? "extension $ext "
                                         : 'Primary Data-Unit ')
                       .'with '.($size/2880)." blocks of non-table data\n"),
                   &skip_data(*IN,$size);
                   next;
    }

    &val_strip(\%vals);

    $vals{XTENSION} !~ /^(?:BINTABLE|TABLE)$/ and
      warn("Skipping unsupported table(".tell(IN).":$size): $vals{XTENSION}\n"),
      &skip_data(*IN,$size),
      next;

    local($isbin) = $vals{XTENSION} eq 'BINTABLE';

    $vals{NAXIS} == 2 or warn("Skipping unsupported $vals{NAXIS}-D tables\n"),
      &skip_data(*IN,$size),
      next;

    $vals{PCOUNT} == 0 or warn("Only supported: Parameter count always 0\n"),
      &skip_data(*IN,$size),
      next;

    $vals{GCOUNT} == 1 or warn("Only supported: Group count always 1\n"),
      &skip_data(*IN,$size),
      next;

    $vals{TFIELDS} > 0 or warn("Unsupported zero-valued TFIELDS\n"),
      &skip_data(*IN,$size),
      next;

    $vals{BITPIX} == 8 or warn("Unsupported BITPIX: $vals{BITPIX}\n"),
      &skip_data(*IN,$size),
      next;

    local($octets);
    local(@tformw);
    local(@tformt);
    if ($isbin) {

      for $i (1..$vals{TFIELDS}) {
        defined($vals{"TFORM$i"}) or warn("Missing mandatory TFORM$i\n"),
          &skip_data(*IN,$size),
          next;
        $vals{"TFORM$i"} =~ /^(\d*)([$qfits::mtform])/o or
          warn("Illegal TFORM$i: $vals{'TFORM'.$i}\n"),
          &skip_data(*IN,$size),
          next;
        $tformt[$i] = length($1) ? $1 : 1; # May be zero;
        $tformw[$i] = $2;
      }

      @length = map($qfits::tformpwidth0{$tformw[$_]} + $qfits::tformpwidth{$tformw[$_]}*$tformt[$_], 1..$vals{TFIELDS});
    } else {
      @length = map($vals{"TBCOL$_"}, 1..$vals{TFIELDS});
    }
    $head && print <<EOF;
#
# file           $file
# extensions     $mext
# --------------------------------------------
# XTENSION       $ext
# Number of    rows $vals{NAXIS2}
# Number of columns $vals{TFIELDS}
#
EOF

    # Entries in first line in output
    @head = map($vals{$_},sort grep(/TTYPE\d/,keys %vals));
    if ($isbin) {
      #Adjust field-sizes
      @l = map(length(),@head);
      grep((!defined($length[$_]) || $l[$_] > $length[$_] and $length[$_] = $l[$_]), 0..$#l);
    }

    $head && print join($sep,map(sprintf('%-'.$length[$_].'s',$head[$_]),0..$#length)),"\n";

    local($filepos) = tell(IN);

    local($_);
    local(@value);
    local($iserr) = 0;
    for $irow (1..$vals{NAXIS2}) {
      @row=();
      local($noctets) = $vals{NAXIS1};
      for $icol (1..$vals{TFIELDS}) {
        if ($isbin) {
          local(@iform)=();
          @value=();
          local($fwidth) = $tformt[$icol] * $qfits::tform{$tformw[$icol]};
          $noctets -= $fwidth;
          if ($tformw[$icol] eq 'A' || $tformw[$icol] eq 'L') {
              local($_);
              read(IN,$_,$fwidth) and @value = map("'$_'", unpack('A'.$tformt[$icol],$_)) or
                &skip_data(*IN,$size-(tell(IN)-$filepos)), $iserr=1, last;
          } elsif ($tformw[$icol] eq 'X') {
              local($_);
              read(IN,$_,$fwidth) and @value = unpack('b'.$tformt[$icol],$_) or
                &skip_data(*IN,$size-(tell(IN)-$filepos)), $iserr=1, last;
          } else {
              local($bitpix) =  $qfits::tformbitpix{$tformw[$icol]};
              read(IN,$_,$fwidth) or
                &skip_data(*IN,$size-(tell(IN)-$filepos)), $iserr=1, last;

              $bitpix != 8 && $qfits::doswap and $_ = &swap($_, abs($bitpix)>>3);

              @value = unpack($qfits::tformunpack{$tformw[$icol]}
                                x $tformt[$icol], $_);
          }
          push(@iform, @value);
          $iserr and last;
          if ($tformw[$icol] eq 'C' || $tformw[$icol] eq 'M') {
            local(@pair)=();
            push(@pair,'('.shift(@iform).','.shift(@iform).')') while (@iform);
            $_ = join(',',@pair);
          } else {
            $_ = join(', ',@iform);
          }
          local($pad) = $length[$icol-1] - length();
          $pad > 0 and $_ .= ' ' x $pad or
          $pad < 0 and $length[$icol-1] -= $pad;
        } else {
          eof(IN) && die "eof($file): $irow $icol";
          $octets = $length[$icol-1];
          read(IN,$_,$octets) == $octets || die "read($file): $irow $icol";
        }
        push(@row, $_);
      }
      $iserr and last;
      $isbin and $noctets and
        warn("Formula 8.2 inconsistency($vals{NAXIS1}:$vals{TFIELDS}) [B]: $noctets\n"),
          &skip_data(*IN,$size-(tell(IN)-$filepos)),
          next;
      print join($sep,@row),"\n";
    }
  }
}

$verbose && warn(&run_stats);
