#!/usr/bin/perl -w

$version = '0.1.1';

# Include-file resides in same directory as program itself
push(@INC,$0 =~ /(.*\/)/);

require 'qfits.pl';

defined($qfits::version) || die;
defined($qfits::support) || die;

$help = "
  $qfits::prog version $version (using qfits.pl version $qfits::version).
  Set the specified FITS-keyword to the specified value in the specified files.

  Proto-type: Options may change in future versions.

Usage:

  $qfits::prog [options] 'FITS KEY NAME' *.fits

A FITS key with spaces must be enclosed in quotes. A space may alternatively be
specified with a '.' (dot). If the key is not already present in the header a
new card will be added. This may increase the size of the fits file with 2880
bytes for each extended header. For FITS keywords that are non-unique (COMMENT,
HISTORY and '        ') a new card is always added.

Options:
  -x  n      : Edit FITS-header extension n. n = 0 means all
               headers. Optional.
  -value     : The value. (Not all fits cards require a value.)
               Special characters like whitespace must be
               enclosed in quotes.
  -comment   : Optional comment. If no comment is specified an
               existing comment is kept (possible truncated).
               If no comment is specified and no comment is
               present, the comment '$qfits::prog set this' is
               added. The value and comment are separated by
               a space, a slash and a space. The comment is
               not inserted if the header already has a FITS
               card with the specified key and value.
  -verbose   : Print some extra information to stderr. [FALSE]
  -help      : Print this message to stderr and exit. [FALSE]

Example:

  $qfits::prog -comment 'For the DO' -value SPEC_CAL_LINES 'HIERARCH ESO PRO CATG'  my*.tfits

For further support contact $qfits::support

";

@ARGV || die $help;

$doext = -1;
$verbose = 0;
$value = $comment = '';
for (;;) {
  $ARGV[0] eq '-value' and &verify_value($ARGV[1]) ||
     die("Invalid value: $ARGV[1]\n".$help),
     $value = $ARGV[1], shift(), shift(), next;
  $ARGV[0] eq '-x' && $ARGV[1] =~ /^\d+$/ and
     $doext = $ARGV[1], shift(), shift(), next;
  $ARGV[0] eq '-comment' and $comment = $ARGV[1], shift(), shift(), next;
  $ARGV[0] eq '-verbose' and shift(), $verbose = 1, next;
  $ARGV[0] eq '-help' and warn($help), exit;
  last;
}

@ARGV || die "Missing key\n".$help;

$key = shift();
$key =~ s/\./ /g;
&verify_key($key) || die("Invalid key: $key\n".$help),

&chk_length($key, $value, $comment);
$comment ||= "$qfits::prog set this";

@ARGV || die $help;

for $file (@ARGV) {

  open(IN,"+<$file") or warn("error: cannot open file [$file]\n"), next;

  local(@expand)=();

  # Count number of extensions
  for ($mext = 0; ; $mext++) {

    if ($doext == 0 || $doext == $mext || $mext == 0) {

      local($headpos) = tell(IN);
      &header(*IN,\@cards) || last;
      local($nexthdu) = tell(IN);

      if ($mext == 0) {
        $cards[0] =~ /^SIMPLE  =/ or warn("Not a FITS file\n"), last;
      } else {
        $cards[0] =~ /^XTENSION=/ or warn("Not a FITS file\n"), last;
      }

      %comments = ();
      &split_cards(\@cards,\%vals,\%comments);

      if ($doext == 0 || $doext == $mext || $mext == 0 && $doext < 0) {
        # Process only the specified extension(s)

        local($doio) = 1;
        if (defined($vals{$key}) && $key !~ /^(?:COMMENT|HISTORY| {8})/) {
          if ($vals{$key} eq $value) {
            $doio = 0;
	  } else {
            for (@cards) {
              if (/^$key\s*(?:=|$|\/)/) {
                local($comment) = defined($comments{$key}) ? $comments{$key}
                                                           : $comment;
                $_ = &card($key, $value, $comment);
                last;
	      }
            }
	  }
        } elsif ($#cards % 36 < 35) {
          local($end)=pop(@cards);
          push(@cards, &card($key, $value, $comment), $end);
        } else {
          push(@expand, tell(IN));
          $doio = 0;
        }
        if ($doio) {
          $verbose && grep(/^$key\s*(?:=|$|\/)/ && warn("$_\n"),@cards);

          seek(IN, $headpos, 0) || die "Could not seek to position $headpos\n";
          print IN @cards;
          seek(IN, $nexthdu, 0) || die "Could not seek to position $nexthdu\n";

        }

        $doext && last;
      }

      $mext==0 and $doext < 0 || !defined($vals{EXTEND}) || $vals{EXTEND} ne 'T'
        and last;

      &skip_data(*IN, &data_size(\%vals)) || last;

    } else {
      &skip_xt_hdu(*IN) || last;
    }

    eof(IN) && last;

  }

  local($nsize) = 2880 * scalar @expand or next;

  eof(IN) or seek(IN,0,2) or
    warn("Could not seek to end of $file, filepos=".tell(IN)." ($nsize)\n"), next;

  # Extend header(s) and insert new card(s)
  local($headpos);
  local($pos)=tell(IN);
  while ($headpos = pop(@expand)) {
    $verbose && warn("Extending ${file}'s header ending at $headpos with $nsize\n");
    while ($pos > $headpos) {
      local($_);
      # This will move the filepos back $nsize
      #  - while moving the block(s) forward by $nsize
      seek(IN,-$nsize,1) && read(IN,$_,$nsize) && print(IN) &&
        seek(IN,-2*$nsize,1) || die "$file $headpos";
      $pos -= $nsize;
    }
    # Filepos for new card here
    seek(IN,$nsize-80-2880,1);
    # Write new card and a hole block with the END-card
    print(IN &card($key, $value, $comment), &card('END'), ' ' x (2880 - 80));
    # Move filepos to old end of header
    seek(IN,-$nsize,1) || die "$file $headpos";
    $verbose && warn("filepos=$pos == $headpos ($nsize)\n");
    $nsize -= 2880;
  }
}

$verbose && warn(&run_stats);


sub is_string {
  local($_)=@_;

  !(/^[TF]$/ || /^(?:$qfits::mfloat)$/ || /^(?:$qfits::mfloat)\s*,\s*(?:$qfits::mfloat)$/);

}

sub verify_key {
  local($_)=@_;

  /^[A-Z_\-][A-Z_\d\-]{0,7}$/ ||
  /^HIERARCH [A-Z_\d\-]+(?: [A-Z_\d\-]+)*/ && length() <= 80;

}


sub verify_value {
  local($_)=@_;

  /^[\040-\176]+$/;

}

sub chk_length {
  @_ = ($key, $value, $comment);

  local($l) = 80;

  if (length($value)) {
    $l -= 2 + length($value);
    $l < 8 and die "Value too long".$help;
  }

  if (length($key) > $l) {
    if (length($value)) {
      die "Key and value too long".$help;
    } else {
      die "Key too long".$help;
    }
  }
}
