
use Config;

if ($Config{byteorder} =~ /^(?:8765)?4321$/) {
  $qfits::doswap = 0;
} elsif ($Config{byteorder} =~ /^1234(?:5678)?$/) {
  $qfits::doswap = 1;
} else {
  die "Unsupported byte order: $Config{byteorder}";
}

# RH LANG problem
$^O eq 'linux' && defined($ENV{LANG}) && $ENV{LANG} =~ /\bUTF-8\b/i and
 die "LANG may not be UTF-8 (try en_US.ISO_8859-1)";

$qfits::version = '0.1';

($qfits::prog) = $0 =~ /([^\/]+)$/;

$qfits::support = 'llundin@eso.org';

$qfits::mkey = '[A-Z\d_\-]{1,8}';

$qfits::mint = '[\-\+]?\d+';
$qfits::mfloat = '(?:'.$qfits::mint.'(?:\.\d*)?(?:[ED]'.$qfits::mint
                .')?|[\-\+]?\.\d+(?:[ED]'.$qfits::mint.')?)';

# A FITS-header consists of 36 cards each containing 80 octets (2880 octets)
$qfits::cards = 'a80' x 36;

# ESO prefixes FITS-keywords with ' '-separated words
#   yielding keywords with a total length in excess of 8
#   Additionally, with prefixing the prefixed word may be longer than 8
#   characters
$qfits::prekey = 'HIERARCH ESO (?:[A-Z\d\-]+ )*[A-Z\d\-]*';

# BINTABLE TFORM data type octet lengths
%qfits::tform = ('L', 1, 'X', 1/8, 'B', 1, 'I', 2, 'J', 4, 'A', 1, 'E', 4, 'D', 8, 'C', 8, 'M', 16, 'P', 8);
%qfits::tformbitpix = ('E', -32, 'D', -64, 'C', -64, 'M', -64);
grep(defined($qfits::tformbitpix{$_}) || ($qfits::tformbitpix{$_} = 8 * $qfits::tform{$_}),
     keys %qfits::tform);
$qfits::mtform = join('', keys %qfits::tform);

%qfits::tformpwidth = ('L', 1, 'X', 1, 'B', 3, 'I', 5, 'J', 9, 'A', 1, 'E', 20, 'D', 20, 'C', 46, 'M', 46, 'P', 9);
%qfits::tformpwidth0 = ('L', 2, 'A', 2);
grep(defined($qfits::tformpwidth0{$_}) || ($qfits::tformpwidth0{$_}=0),
     keys %qfits::tformpwidth);

%qfits::tformunpack=('D', 'd', 'E', 'f', 'C', 'ff', 'M', 'dd', 'J', 'l', 'I', 's', 'B', 'C');

sub data_entries {
  local(*vals)=@_;

  # Compute the number of entries in the Data Unit described by %vals

  !defined($vals{NAXIS}) || $vals{NAXIS} < 0 and
     warn("Missing/Illegal NAXIS card\n"), return;

  if (defined($vals{XTENSION})) {
    # These two are mandatory in extension headers only
    defined($vals{PCOUNT}) or
      warn("Warning: Missing Mandatory PCOUNT card, assuming value 0\n"),
     $vals{PCOUNT} = 0;
    defined($vals{GCOUNT}) or warn("Warning: Missing Mandatory GCOUNT card, assuming value 1\n"),
     $vals{GCOUNT} = 1;
  }

  local($size) = $vals{NAXIS} ? 1 : 0;
  for (1..$vals{NAXIS}) {
    defined($vals{"NAXIS$_"}) or warn("Missing NAXIS$_ card\n"), return;
    $vals{"NAXIS$_"} < 0 and warn("Illegal negative-valued NAXIS$_\n"), return;
    $size *= $vals{"NAXIS$_"};
  }

  if (defined($vals{PCOUNT})) {
    $vals{PCOUNT} < 0 and warn("Illegal negative-valued PCOUNT card\n"), return;
    $size += $vals{PCOUNT};
  }
  if (defined($vals{GCOUNT})) {
    $vals{GCOUNT} < 0 and warn("Illegal negative-valued GCOUNT card\n"), return;
    $size *= $vals{GCOUNT};
  }

  $size;

}

sub data_size {
  local(*vals)=@_;

  # Compute the size of the Data Unit

  defined($vals{BITPIX}) or warn("Missing BITPIX card\n"), return;
  $vals{BITPIX} ==   8 || $vals{BITPIX} == 16 || $vals{BITPIX} == 32 ||
  $vals{BITPIX} == -64 || $vals{BITPIX} == -32 or
    warn("Illegal BITPIX card: $vals{BITPIX}\n"), return;

  local($size) = &data_entries(\%vals) or return;

  $size *= abs($vals{BITPIX})>>3;

  # The size of a data-unit is a multiplum of 2880 octets
  $size % 2880 and $size += 2880 - $size % 2880;

  $size;

}

sub skip_xt_hdu {
  local(*IN)=@_;

  # Advance file-pointer from start of extension header to the start of the next one

  # Returns 1 on success, undef on error

  # First find keywords that determine size of data-segment
  local($_);
  local($ok) = 0;
  local($found) = 0;
  local(%vals);

  for (; !$ok; ) {
    read(IN,$_,2880) == 2880 or eof(IN) && seek(IN, 0, 2),
      warn("Truncated header record at filepos=".tell(IN)."\n"), return;

    for (unpack($qfits::cards,$_)) {
      if (/^(NAXIS\d*|BITPIX|PCOUNT|GCOUNT|XTENSION)\s*=\s*(\S+)/) {
        $vals{$1} = $2;
        $found++;
        $ok = defined($vals{NAXIS}) && $found == 5 + $vals{NAXIS} and last;
      }
    }
    $ok or /END {77}(?: {80})*$/ and eof(IN) && seek(IN, 0, 2),
      warn("Missing mandatory keywords in extension header, filepos=".tell(IN)."\n"),
      return;
  }

  # Make sure that any remaining header records are read
  for (;!/END {77}(?: {80})*$/;) {
    read(IN,$_,2880) == 2880 or eof(IN) && seek(IN, 0, 2),
      warn("Truncated header record at filepos=".tell(IN)."\n"), return;
  }

  &skip_data(*IN, &data_size(\%vals));

}


sub skip_header {
  local(*IN)=@_;

  # Advance the file-pointer from the begining of a header unit to just beyond it
  local($_);

  for (;;) {
    read(IN, $_, 2880) == 2880 or eof(IN) && seek(IN, 0, 2),
      warn("Truncated header record at filepos=".tell(IN)."\n"), last;

    /END {77}(?: {80})*$/ && last;

  }

}

sub skip_data {
  local(*IN,$size)=@_;

  # Advance the file-pointer by $size octets

  defined($size) || return;
  $size == 0 && return 1;

  local($seek) = tell(IN);
  seek(IN, $size-1, 1);

  # IN may be STDIN, in which case seek() seems to not work
  # Read the data instead :-(
  # Even when the input is not STDIN, seek may return 1 ("success"),
  # even though EOF would have been hit with a smaller $size
  # Therefore seek 1 octet less, and read the last octet.

  $seek += $size - tell(IN);
  local($buf);
  if ($seek % 2880) {
    $seek -= read(IN, $buf, $seek % 2880);
    $seek % 2880 and
      eof(IN) && seek(IN, 0, 2),
      warn("Data unit (${size}B) skip failed, filepos=".tell(IN)."\n"),
      return undef;
  }

  if ($seek) {
    1 while (read(IN,$buf,2880) == 2880 and $seek -= 2880);
  }

  $seek == 0 or eof(IN) && seek(IN, 0, 2),
    warn("Data unit (${size}B) skip failed, Filepos=".tell(IN)."\n"),
    undef;

}

sub header {
  local(*IN,*cards)=@_;

  # Read a FITS-header and advance the file-pointer to just beyond the header

  # Returns 1 on success, undef on error

  @cards=();

  local($_);
  local($unpack) = $qfits::cards;
  local($end) = 0;
  local($size) = 0;
  for (; !$end; ) {
    local($r) = read(IN,$_,2880);
    defined($r) && $r == 2880 or eof(IN) && seek(IN, 0, 2),
      warn("Invalid header record truncated at filepos=".tell(IN)."\n"), return;
    $size += $r;

    $end = s/(END {77})(?: {80})*$/$1/ and $unpack = 'a80' x (length()/80);

    push(@cards,unpack($unpack,$_));

  }

  $size;
}

sub split_cards {
  local(*cards,*vals,*coms)=@_;

  # Split FITS-header cards into a value- and comment-hash

  %vals=%coms=();
  local($icard) = 0;
  for (@cards) {
    $icard++;
    if (/^((?:$qfits::prekey)?$qfits::mkey)\s*= +( |[TF]|$qfits::mfloat|\(\s*$qfits::mfloat\s*,\s*$qfits::mfloat\s*\)|\'(?:[^\']|\'\')*\')\s*(?:\/ ?(.*))?/o) {
      $vals{$1} = $2;
      $coms{$1} = $3;
    } elsif (/^( {8}|HISTORY|COMMENT)(.+)/) {
      $vals{$1} = $2;
      $coms{$1} = $1;
    } elsif (/^(?:END|   ) {77}$/) {
      1;
    } else {
      warn("Unparsable FITS card($icard:".length()."): $_\n");
    }
  }

}

sub val_strip {
  local(*vals)=@_;

  # Strip quotes and trailing space from string-values in a FIT-header

  grep($vals{$_} =~ s/^\'(.*?)\s*\'$/$1/ && $vals{$_} =~ s/\'\'/\'/g, keys %vals);

}

sub swap {
  local($_,$octets)=@_;

  # Endian-swap a binary record

  # returns undef on error

  $octets == 1 && return $_;

  $octets == 2 || $octets == 4 || $octets == 8 or
    warn("Cannot Endian-swap sequences of $octets octets\n"), return;

  length() % $octets and
    warn("Cannot Endian-swap a sequence of length ".length()."\n"), return;

  local($n) = length()/$octets;

  if ($octets == 8) {
    # Has to skip forward 4 octets, swap those 4 octets, then back up 8 octets,
    # then swap those first 4 octets, then skip forward 4 to the start of the
    # next 8 octets
    pack('NN' x $n, unpack('x4VX8Vx4' x $n, $_));
  } elsif ($octets == 4) {
    pack('N' x $n, unpack('V' x $n, $_));
  } else {
    pack('n' x $n, unpack('v' x $n, $_));
  }

}

sub read_data_record {
  local(*IN,*nums,$bitpix)=@_;

  # Read a data record (2880 octets) and convert it according to bitpix:
  # -64:  360 doubles
  # -32:  720 floats
  #  32:  720 32-bit signed integers
  #  16: 1440 16-bit signed integers
  #   8: 2880  8-bit unsigned integers

  # Don't use this for BITPIX=8 character data, instead use
  #   read(IN,$_,2880) and unpack('a' x 2880,$_)
  # For BINTABLE data use bitpix = -8*$vals{NAXIS1} / $vals{TFIELDS}, ie.
  #   -64 or -32

  # It is assumed that the 'd'/'f'-unpack() of perl returns IEEE-754 format.

  local($_);
  read(IN,$_,2880) == 2880 or eof(IN) && seek(IN, 0, 2),
     warn("Failed read at filepos ".tell(IN)."\n"),
    return;

  $bitpix != 8 && $qfits::doswap and $_ = &swap($_, abs($bitpix)>>3);

  if ($bitpix == -64) {
    @nums = unpack('d' x 360, $_);
  } elsif ($bitpix == -32) {
    @nums = unpack('f' x 720, $_);
  } elsif ($bitpix == 32) {
    @nums = unpack('l' x 720, $_);
  } elsif ($bitpix == 16) {
    @nums = unpack('s' x 1440, $_);
  } elsif ($bitpix == 8) {
    @nums = unpack('C' x 2880, $_);
  } elsif ($bitpix < 0) {
    warn("Unsupported BITPIX for IEEE-754 unpack(): $bitpix\n");
    return;
  } else {
    warn("Unsupported BITPIX for integer unpack(): $bitpix\n");
    return;
  }
  scalar @nums;
}

sub card {
  local($keyname,$value,$comment)=@_;

  # Format a (key,val,comment)-triplet for output to a FITS-header

  if ($keyname eq 'END') {
    sprintf("%-80s",$keyname);
  } elsif ($keyname =~ /^(?: {8}|HISTORY|COMMENT)$/) {
    length($value || $comment) > 72 &&
      warn("Truncating too long value in $keyname-card: ".($value || $comment)."\n");
    substr(sprintf("%-8s%-72s",$keyname,$value || $comment),0,80);
  } else {
    local($l) = 0;
    $value =~ /^(?:[TF]|$qfits::mfloat|\(\s*$qfits::mfloat\s*,\s*$qfits::mfloat\s*\))$/o or
      $l = 1,
      $value =~ s/\'/\'\'/g, # O'HARA -> O''HARA
      $value = sprintf("'%-8s'",$value);
    local($_) = sprintf("%-8s= %s",$keyname,$value);
    if (length() > 80) {
      warn("Truncating too long value in $keyname-card: $value / $comment\n");
      $_ = substr($_,0,80);
      $l and s/.(\'*)$/\'/; # Add end-quote
      $_;
    } elsif ($comment) {
      local($pad) = 80 - 3 - length();
      $_ .= sprintf(" / %${pad}s",$comment);
      if (length() > 80) {
        warn("Truncating too long comment in $keyname-card: $value / $comment\n");
        substr($_,0,80);
      } else {
        sprintf("%-80s",$_);
      }
    }
  }
}

sub run_stats {

 # Prepare a string with CPU and memory usage

  local($m) = 0;

  # Get memory usage from /proc/
  if (open(IN,"/proc/$$/statm")) {

    if (<IN> =~ /(\d+)\s+(\d+)/) {
      # [$2] == 4KB
      $m = $2;
      $^O eq 'linux' && $1 > $m and $m=$1;
    }

    close(IN);
 }

 sprintf("Wall-clock [s]: %1d, CPU [s]: %1d, System [s]: %1d",time-$^T,(times)[0..1])
 .($m && sprintf(", Memory [kB]: ".(4*$m)))."\n";

}

1;
