#!/usr/bin/perl -w

$version = '0.1.1';

# Include-file resides in same directory as program itself
push(@INC,$0 =~ /(.*\/)/);

require 'qfits.pl';

defined($qfits::version) || die;
defined($qfits::doswap) || die;
defined($qfits::support) || die;

$help = "
  $qfits::prog version $version (using qfits.pl version $qfits::version).
  Convert a text-file with numbers in columns from stdin to a FITS-file
  with a binary table extension to stdout. Lines starting with something
  else than a number (and any preceeding whitespace) are ignored.

  Proto-type: Options may change in future versions.

Usage:

  $qfits::prog [options] col_name1 [col_name2 [col_name3 [...]]]

Options:
  -using col1[:col2[:col3]]]  : Specify which input columns to use.
                                Numbering starts with 1. The number of columns
                                specified with this option must equal the number
                                of column names. This option supports gnuplot-
                                style, eg. -using '1e-6*\$1:\$2-\$3', such usage
                                will increase the CPU-usage. Absence of -using
                                means all input columns are used, this reduces
                                CPU-usage.
  -rows nrows                 : Specify the number of rows in the output. The
                                input must contain at least this many rows.
                                Absence of -rows means all input rows, this
                                necessitates storage of the entire data unit in
                                memory.
  -reverse                    : Reverse the order of the table rows.
                                Necessitates storage of the entire data unit in
                                memory. [FALSE]
  -verbose                    : Print some extra information to stderr. [FALSE]
  -help                       : Print this message to stderr and exit. [FALSE]

Example:

  $qfits::prog -rows `wc -l < myhuge.txt` colm{1,2,3} < myhuge.txt > myhuge.tfits
  $qfits::prog -using 3:2 -reverse 'Wavelength [m]' Intensity < my.txt > my.tfits

For further support contact $qfits::support

";

@ARGV || die $help;

$mrows = 0;
$reverse = 0;
$verbose = 0;
for (;;) {
  $ARGV[0] eq '-using' and $ARGV[1] =~ /^([^\:]+(?:\:[^\:]+)*)$/ || die($help),
    $using = $1, shift(), shift(), next;
  $ARGV[0] eq '-rows' and $ARGV[1] =~ /^(\d+)$/ && $1 > 0 || die($help),
    $mrows = $1, shift(), shift(), next;
  $ARGV[0] eq '-reverse' and $reverse = 1, shift(), next;
  $ARGV[0] eq '-verbose' and shift(), $verbose = 1, next;
  $ARGV[0] eq '-help' and warn($help), exit;
  last;
}

@cols = @ARGV;
$ncols = scalar @cols;


@using = ();
$need_eval = 0;
if (defined($using)) {
  @using = split(/:/,$using);

  local($maxcol) = 0;

  # See if conversion can be done without the expensive eval()
  $need_eval = grep(!/^\$?[1-9]\d*$/, @using);

  if ($need_eval) {
    # Convert pure digits and '$digits'-constructs to $_[digits-1]...
    grep(s/^([1-9]\d*)$|\$([1-9]\d*)/&use_conv/eg, @using);

    local(@_) = grep(1,1..$maxcol);
    grep(eval(),@using);
    $@ and die "Invalid -using: $@\n".$help;

  } else {
    # @using only specifies the columns and their order
    grep(s/^\$?([1-9]\d*)$/$1-1/e, @using) == $#using + 1 or
      die join(' ',"Invalid -using:",@using)."\n$help";

    # Find largest input columns number
    grep($_ > $maxcol && ($maxcol = $_),@using);
    $maxcol++;

  }

  $maxcol || die "-using do not specify any input columns\n".$help;
  $verbose and warn("Expecting at least $maxcol columns in input\n");

}

$ncols and !defined($using) || scalar @using == $ncols or die $help;

$octets = 8;

$naxis1 = $octets * $ncols;

# Pack the values one FITS-record (2280 octets = 360 doubles) at a time
$nd = 2880/$octets;

if ($reverse) {

  # Reverse input, cannot print any data-unit records before the last row is read
  local($buf)='';
  local($bsize) = 0;
  for ($nrows = 0; $mrows == 0 || $nrows < $mrows and $_ = <STDIN>; $nrows++) {
    /^\s*[\-\+]?\.?\d/o or $nrows--, next;

    # Store the rows in reverse order
    unshift(@values, &line2row($_));

    $#values >= $nd-1 || next;

    $_ = pack('d' x $nd, splice(@values,0,$nd));

    # Appending is much faster than prepending, needs to print the records
    # in reverse order
    $buf .= $qfits::doswap ? &swap($_,$octets) : $_;
    $bsize += 2880;

  }

  $mrows == 0 || $nrows == $mrows or die "Input has only $nrows rows\n".$help;

  print &table_headers($nrows,$naxis1,@cols);

  $nd= 1 + $#values;
  if ($nd> 0) {

    $_ = pack('d' x $nd, @values);

    print $qfits::doswap ? &swap($_,$octets) : $_;

  }

  # Print the records in reverse order
  $bsize-=2880, print substr($buf,$bsize,2880) while ($bsize > 0);

} elsif (!$mrows) {

  # Need to keep the entire data-unit in memory, because rows must be counted
  local($buf)='';
  for ($nrows = 0; <STDIN>; $nrows++) {
    /^\s*[\-\+]?\.?\d/o or $nrows--, next;

    push(@values, &line2row($_));

    $#values >= $nd-1 || next;

    $_ = pack('d' x $nd, splice(@values,0,$nd));

    $buf .= $qfits::doswap ? &swap($_,$octets) : $_;

  }

  $nrows or die $help;

  print &table_headers($nrows,$naxis1,@cols);
  print $buf;

  $nd= 1 + $#values;
  if ($nd> 0) {

    $_ = pack('d' x $nd, @values);

    print $qfits::doswap ? &swap($_,$octets) : $_;

  }

} else {

  # Number of rows known, header and data units records can be printed
  # as soon as they are generated

  print &table_headers($mrows,$naxis1,@cols);

  for ($nrows = 0; $nrows < $mrows and $_ = <STDIN>; $nrows++) {
    /^\s*[\-\+]?\.?\d/o or $nrows--, next;

    push(@values, &line2row($_));

    $#values >= $nd-1 || next;

    $_ = pack('d' x $nd, splice(@values,0,$nd));

    print $qfits::doswap ? &swap($_,$octets) : $_;

  }

  $nrows == $mrows or die "Input has only $nrows rows\n".$help;

  $nd= 1 + $#values;
  if ($nd> 0) {

    $_ = pack('d' x $nd, @values);

    print $qfits::doswap ? &swap($_,$octets) : $_;

  }

}

# Pad data unit with spaces
$nd && $nd< 360 and print ' ' x (2880 - $nd*$octets);

$verbose && warn(&run_stats);

sub use_conv {

  $+ > $maxcol and $maxcol = $+;

  sprintf('$_[%1d]',$+-1);

}


sub line2row {
  local($_)=@_;

  # If using is not defined simply use all fields
  defined($using) || return split();

  local(@_) = split();

  $need_eval ?  map(eval(),@using) : map($_[$_],@using);

}

sub table_headers {
  local($nrows,$naxis1,@cols)=@_;

  local($ncols) = scalar @cols;

  local($ENV{TZ}) = 'GMT';

  local($s,$m,$h,$mday,$mon,$y)=localtime();
  $y += 1900;
  $mon++;

  local($utc) = sprintf("%4d-%02d-%02dT%02d:%02d:%02d",$y,$mon,$mday,$h,$m,$s);

  # A BINTABLE FITS-extension-header
  local($bintable) =
    &card('XTENSION','BINTABLE','FITS Binary Table Extension').
    &card('BITPIX',  8,         '8-bit character format').
    &card('NAXIS',   2,         'Tables are 2-D char. array').
    &card('NAXIS1',  $naxis1,   'Bytes in a row').
    &card('NAXIS2',  $nrows,    'No. of rows in the table').
    &card('PCOUNT',  0,         'Parameter count always 0').
    &card('GCOUNT',  1,         'Group count always 1').
    &card('TFIELDS', $ncols,    'No. of cols in the table');

  for $i (1..$ncols) {
    $bintable .= &card("TFORM$i",'1D','Format of field');
    $bintable .= &card("TTYPE$i",$cols[$i-1],'Field label');
    $cols[$i-1] =~ /\[([ a-z\*\/\^\-\d])+\]/ and
    $bintable .= &card("TUNIT$i",$1,'Physical unit of field');
  }


  # A space-padded minimal FITS-header
  sprintf("%-2880s",
    &card('SIMPLE', 'T','Standard FITS file').
    &card('BITPIX', 8,  'ASCII or bytes array').
    &card('NAXIS',  0,  'Minimal header').
    &card('EXTEND', 'T','There is a FITS extension').
    &card('BLOCKED','T','The file may be blocked').
    &card('END')).

  # - printed with space-padding

  sprintf("%-2880s", $bintable.
    &card('ORIGIN','ESO-QFITS','Written by qfits.pl').
    &card('DATE',  $utc,       '[UTC] Date of writing').
    &card('END'));

}
